// -*- C++ -*-
// $RCSfile: progressbar.C,v $
// $Revision: 1.21 $
// $Author: langer $
// $Date: 2005/12/13 16:45:19 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/IO/progressbar.h"
#include "common/lock.h"
#include "common/threadstate.h"
#include "common/tostring.h"
#include <iostream>
#include <pthread.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <vector>
#include <iostream>


std::vector<Progress*> &listOfProgress() {
  static std::vector<Progress*> listPbars;
  return listPbars;
}

Progress *getProgress() {
  // Default ThreadID constructor makes one that identifies the current thread.
  ThreadID currentThread; 

  std::vector<Progress*> &list = listOfProgress();
  for(std::vector<Progress*>::size_type i=0; i<list.size(); i++)
    if(list[i]->get_thread_ID() == currentThread)
      {
	return list[i];
      }
  return new Progress();
}



bool operator==(Progress &p1, Progress &p2) {
  return p1.get_threadstate() == p2.get_threadstate();
}


Progress::Progress()
  : progress(0),
    stop(0),
    success(INCOMPLETE),
    bar_msg(" ")
{

  tState = findThreadState();
  listOfProgress().push_back(this);
}




Progress::~Progress() {
  remove_from_list(); //if still on list, remove yourself
}

void Progress::reset() {
  success = INCOMPLETE;
  stop = 0;
  progress = 0;
  bar_msg = " ";
}

void Progress::stop_it() {
  stop = 1;
}

void Progress::set_success() {
  success = COMPLETE;
}

void Progress::set_failure() {
  success = FAILED;
}

ProgressBarStatus Progress::get_success() const {
  return success;
}

float Progress::get_progress()
{
  KeyHolder thekey(lockobj);
  return progress;  
}

ThreadID &Progress::get_thread_ID() const {
  return tState->get_thread_ID();
}

void Progress::set_message(const std::string & msg)
{
  KeyHolder thekey(lockobj);
  bar_msg = msg;
}

const std::string Progress::get_message() {
  lock();
  std::string temp_bar_messg = bar_msg;
  unlock();
  return temp_bar_messg;
}

void Progress::remove_from_list()
{
  std::vector<Progress*> &list = listOfProgress();
  for(std::vector<Progress*>::iterator i=list.begin(); i<list.end(); ++i)
    if(*i == this)
	list.erase(i);
}

void Progress::cleanup() {
  remove_from_list();
}

void Progress::make_defunct() {
  tState->make_defunct();
}
