# -*- python -*-
# $RCSfile: utils.py,v $
# $Revision: 1.54 $
# $Author: reida $
# $Date: 2006/07/26 18:05:58 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import lock
from oof2.common import debug
import string
import sys
import types

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to make a tuple out of its inputs.  This is useful
# because sometimes we want to use the same loop to iterate over
# things, even if the things are not in an indexable object.
#   This function makes no effort to tupify sublists or other
# internal components of the passed-in object.
def tupify(x):
    try:
        return tuple(x)
    except (TypeError, AttributeError):
        return (x,)
    
def listify(x):
    try:
        return list(x)
    except (TypeError, AttributeError):
        return [x]

# Argument is presumptively either a string or a list of strings.
# This is needed to pass around the profiles for the boundary
# conditions.
def string_listify(x):
    if type(x) == type(''):
        return [x]
    else:
        return x

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to return adjacent pairs of a list, i.e. given [1,2,3],
# it returns [(1,2), (2,3)].
def list_pairs(lst):
    return [ (lst[i], lst[i+1]) for i in range(len(lst)-1) ]

# Possible to write in terms of generators, if you're sure that
# nobody's going to try to use Python 2.1 or earlier.
# from __future__ import generators
# def list_pairs(lst):
#     for i in range(len(lst)-1):
#         yield (lst[i],lst[i+1])

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
# Function to return 1 if the argument is indexable and 0 otherwise
def isIndexable(x):
    try:
        len(x) # I suppose one could imagine python optimizing this away!
        return 1
    except TypeError:
        return 0
    except AttributeError:
        return 0

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
# Used by boundary-condition consistency check to see if small tuples
# overlap.  Returns a possibly-empty tuple of the intersecting
# components.
def tupleIntersect(x,y):
    result = ()  # Empty tuple.
    for xvalue in x:
        for yvalue in y:
            if xvalue == yvalue:
                result+=(xvalue,)
    return result

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to take a list of lists and flatten it into a single list.
def flatten(lols):
    result = []
    for l in lols:
        result += l
    return result

# Unfortunately, this version is much slower:
##   def flatten(lols): return [x for y in lols for x in y]

def flatten_all(lols):
    result = []
    for l in lols:
        if type(l) is types.ListType or type(l) is types.TupleType:
            result += flatten_all(l)
        else:
            result.append(l)
    return result

# Given a template which is a list of lists of lists to arbitrary
# depth, convert the flat datalist to a nested list with the same
# structure.  Each sublist of the result will have the same length as
# the corresponding sublist in the template.
def unflatten(template, datalist):
    which = 0
    result, which = _do_unflatten(template, datalist, which)
    return result

def _do_unflatten(template, datalist, which):
    result = []
    for obj in template:
        if type(obj) == types.ListType or type(obj) == types.TupleType:
            sublist, which = _do_unflatten(obj, datalist, which)
            result.append(sublist)
        else:
            result.append(datalist[which])
            which += 1
    return result, which

## flatten1 takes a list of lists and returns a list of the contents
## of all the sublist.  It *doesn't* do any recursion on the sublists.
## Creating a list of lists and calling flatten1 on it is much more
## efficient than creating the list by calling 'list += sublist' in a
## loop.

def flatten1(lols):
    length = reduce(lambda x,y: x+len(y), lols, 0)
    flatlist = [None]*length
    offset = 0
    for lst in lols:
        flatlist[offset:offset+len(lst)] = lst
        offset += len(lst)
    return flatlist


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Get a list of the classes to which an object or class belongs.
def classes(c):
    if type(c) == types.InstanceType:
        return classes(c.__class__)
    if type(c) == types.ClassType:
        if not c.__bases__:
            return [c]
        return [c] +  flatten(map(classes, c.__bases__))
    return []                       # not a class or an instance of a class

# Get the *names* of the classes to which an object or class belongs.
# This loses namespace information, so it's not as robust as using
# classes().  But it's more readable.
def classnames(c):
    return [cl.__name__ for cl in classes(c)]

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Replaced by skeletonnode.canonicalorder, which queries indices.
# def canonicalorder(p0, p1):
#     if p0 < p1:
#         return (p0, p1)
#     return (p1, p0)

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Evaluate expressions, statements and files in the main oof namespace.

mainmodule = sys.modules['__main__']

def OOFexec(command):
    exec(command, mainmodule.__dict__)

def OOFeval(expr):
    return eval(expr, mainmodule.__dict__)

def OOFexecfile(file):
    execfile(file, mainmodule.__dict__)

# Define an existing object in the main oof namespace.
def OOFdefine(name, obj):
    mainmodule.__dict__[name] = obj

# Safe, restricted evaluation: can't import or do anything malicious.
# Can only be used to evaluate things defined in the main namespace.
# Is used when loading data files.

def OOFeval_r(expr):
    return mainmodule.__dict__[expr]

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Parse an argument list, e.g. from inside an "eval" when the
# arguments arrive in string form.

    
def argback(*args,**kwargs):
    return {"tuple":args, "dictionary":kwargs}

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Construct a name, based on the given name arg, that does not occur
# in the given list of other names.  If the given name isn't in the
# list, return it.  If it is, append "<number>" to it, where 'number'
# is chosen so that the result is unique.

# The list of names passed to uniqueName is almost always a global
# list, and so in a threaded environment we need to prevent
# simultaneous uniqueName calls with the same name list.  Since the
# calls are quick and not done in time sensitive situations, we put a
# single mutex lock inside uniqueName, instead of requiring all
# callers to implement their own lock.

import re

_uniqueNameLock = lock.SLock()

def uniqueName(name, othernames, exclude=None):
    _uniqueNameLock.acquire()
    others = othernames[:]
    try:
        if exclude is not None:
            try:
                others.remove(exclude)
            except ValueError:
                pass
        if name not in others:
            return name
        # Strip '<number>' suffix, if any.
        basename = re.split('<[0-9]+>$', name)[0]
        # Find any existing names of the form 'basename<number>'.
        expr = re.compile("^" + basename + "<[0-9]+>$")
        matches = filter(expr.match, others)
        if matches:
            # Find largest existing "<number>".
            suffixes = [x[len(basename)+1:-1] for x in matches]
            lastsuffix = max(map(eval, suffixes))
        else:
            lastsuffix = 1
        return "%s<%d>" % (basename, lastsuffix+1)
    finally:
        _uniqueNameLock.release()
        
# Special version that uses underscores, for menu names.
def menUniqueName(name, othernames):
    _uniqueNameLock.acquire()
    try:
        if name not in othernames:
            return name
        # Strip '_number' suffix, if any.
        basename = re.split('_[0-9]+$', name)[0]
        # Find any existing names of the form 'basename_number'.
        expr = re.compile("^" + basename + "_[0-9]+$")
        matches = filter(expr.match, othernames)
        if matches:
            # Find largest existing "number".
            suffixes = [x[len(basename)+1:] for x in matches]
            lastsuffix = max(map(eval, suffixes))
        else:
            lastsuffix = 1
        return "%s_%d" % (basename, lastsuffix+1)
    finally:
        _uniqueNameLock.release()

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Convert a menu item name to the form displayed by the GUI.
# Underscores are replaced by spaces, and double underscores are
# replaced by underscores.  The sleazy implementation here also
# converts double spaces to underscores, but nobody in his right mind
# would put double spaces in menu item names, right?

def underscore2space(name):
    return name.replace("_", " ").replace("  ", "_")

# And the not-quite-inverse.
def space2underscore(name):
    return name.replace(" ", "_")

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#



# Utility class for a list with no duplicates.  Defers uniquification
# until a request for a list member is made.  Provides a full suite of
# list emulation functions.
class UniqueList:
    def __init__(self, list=[]):
        self.datalock = lock.SLock()
        self.data = []
        self.unique = 1 # Default case for empty list, gets zero'd in append.
        for l in list:
            self.append(l)

    def append(self, item):
        self.datalock.acquire()
        try:
            self.data.append(item)
            self.unique = 0
        finally:
            self.datalock.release()

    # Uniquification function -- actual uniquification is deferred
    # until retrieval, so there is no overhead for repeated additions.
    # Uniquifcation does not lock, because it presumes its caller has
    # the lock.
    def uniquify(self):
        if self.unique==0:
            self.data.sort()
            nuniq = 1
            for i in range(1, len(self.data)):
                if self.data[i] != self.data[i-1]:
                    if nuniq < i:
                        self.data[nuniq] = self.data[i]
                    nuniq += 1
            del self.data[nuniq:]
            self.unique = 1
            
    def remove(self, item):
        self.datalock.acquire()
        try:
            self.uniquify()
            try:
                i = self.data.index(item)
            except ValueError:
                pass # Removal of non-entry is harmless, unlike regular lists.
            else:
                del self.data[i]
        finally:
            self.datalock.release()

    def index(self, value):
        self.datalock.acquire()
        try:
            self.uniquify()
            return self.data.index(value)
        finally:
            self.datalock.release()

    # "Add" will only work if "other" is a sequence type.
    # Returns a new object.
    def __add__(self, other):
        try:
            len(other)
        except TypeError:
            raise TypeError(
                "Only sequence types can be appended to UniqueList.")
        else:
            self.datalock.acquire()
            try:
                self.uniquify()
                res = UniqueList()
                res.data = self.data[:]
                for o in other:
                    res.append(o)
                return res
            finally:
                self.datalock.release()

    # Allow UniqueLists to be appended to regular lists -- in
    # such a case, copy out the data references.
    def __radd__(self, other):
        self.datalock.acquire()
        try:
            self.uniquify()
            if type(other)==types.ListType:
                return other + self.data[:]
            raise TypeError(
                "UniqueLists can only be appended to ordinary lists.")
        finally:
            self.datalock.release()
    
    # Insertion is actually meaningless, since UniqueLists get sorted.
    # It's provided for completeness, discards "pos" and calls append.
    # Does not need to lock, since append does.
    def insert(self, pos, value):
        self.append(value)

    # Remaining list emulation functions, except for "__len__", which
    # is in the parent class.
    def __getitem__(self, index):
        self.datalock.acquire()
        try:
            self.uniquify()
            return self.data[index]
        finally:
            self.datalock.release()

    def __getslice__(self, start, end):
        self.datalock.acquire()
        try:
            self.uniquify()
            return self.data[start:end]
        finally:
            self.datalock.release()
    
    def __setitem__(self, index, value):
        self.datalock.acquire()
        try:
            self.uniquify()
            self.data[index] = value
        finally:
            self.datalock.release()

    def __delitem__(self, index):
        self.datalock.acquire()
        try:
            self.uniqiufy()
            del self.data[index]
        finally:
            self.datalock.release()

    def __contains__(self, value):
        self.datalock.acquire()
        try:
            return value in self.data
        finally:
            self.datalock.release()

    def __len__(self):
        self.datalock.acquire()
        try:
            self.uniquify()
            return len(self.data)
        finally:
            self.datalock.release()

    # Equality comparison only permitted with lists or UniqueLists.
    def __eq__(self, other):
        if type(other)!=types.ListType and \
           (type(other)!=types.InstanceType or other.__class__!=UniqueList):
            return False
        self.datalock.acquire()
        try:
            for o in other:
                if o not in self.data:
                    return False
            for s in self.data:
                if s not in other:
                    return False
            return True
        finally:
            self.datalock.release()

    def __ne__(self,other):
        return not self.__eq__(other)

    def __repr__(self):
        self.datalock.acquire()
        try:
            self.uniquify()
            return "UniqueList("+`self.data`+")"
        finally:
            self.datalock.release()

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

def weedlist(condition, liszt):
    nextspot = 0
    for x in liszt:
        if condition(x):
            if liszt[nextspot] is not x:
                liszt[nextspot] = x
            nextspot += 1
    del liszt[nextspot:]
    return liszt

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
def screenwidth():
    try:
        import curses
        scr = curses.initscr()
        width = scr.getmaxyx()[1]
        curses.endwin()
    except ImportError:
        width = 80
    return width

# Function to format a long string so that it fits into strings of
# length width.  Returns a list of strings guaranteed to be shorter
# than width, provided there are some spaces in there somewhere.
def format(line, width):
    linelist = [string.strip(ell) for ell in line.split("\n")]
    outlist = []
    for str in linelist:
        while len(str) > width:
            breakpoint = string.rfind(str," ",0,width)
            outlist.append(str[0:breakpoint])
            str = str[breakpoint+1:]
        else:
            outlist.append(str)
    return outlist


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Minimal ordered dictionary class.  If more functionality is
# required, get someone else's code off the web.  This class just
# ensures that the objects are returned in the order in which they
# were added.

class OrderedDict(dict):
    def __init__(self):
        dict.__init__(self)
        self._keys = []
    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        if key not in self._keys:
            self._keys.append(key)
    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self._keys.remove(key)
    def keys(self):
        return self._keys
    def values(self):
        return [dict.__getitem__(self, key) for key in self._keys]
    def items(self):
        return [(key, dict.__getitem__(self, key)) for key in self._keys]
    def reorder(self, keylist):
        # Make sure that our keys are in the order given by keylist.
        # keylist must contain all of our keys, but may contain more.

        ## Check for keys in self._keys that aren't in keylist
        unlisted = [key for key in self._keys if key not in keylist]

        self._keys = [key for key in keylist if key in self._keys] + unlisted
    def __eq__(self, other):
        return self._keys == other._keys and \
               super(OrderedDict, self).__eq__(other)


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# A list-like object in which space can be reserved so that it doesn't
# have to be reallocated all the time.  It doesn't support all of the
# slice operations that a real list would.

class ReservableList:
    def __init__(self, n=0):
        self._list = [None]*n
        self._length = 0
    def __len__(self):
        return self._length
    def __getitem__(self, i):
        if type(i) is types.IntType:
            if i >= self._length or i < -self._length+1:
                raise IndexError(i, "is out of range")
        return self._list[i]
    def __setitem__(self, i, val):
        if i >= 0:
            self._list[i] = val
        else:
            self._list[self._length+i] = val
    def capacity(self):
        return len(self._list)
    def reserve(self, size):
        if size <= len(self._list):
            return
        new_list = self._list + [None]*(size-len(self._list))
        self._list = new_list
    def reverse(self):
        oldsize = len(self._list)
        newlist = self._list[:self._length]
        newlist.reverse()
        self._list = newlist + [None]*(oldsize - self._length)
    def append(self, x):
        if self._length == len(self._list):
            self._list.append(x)
        else:
            self._list[self._length] = x
        self._length += 1
    def __delitem__(self, i):
        oldsize = len(self._list)
        if type(i) is types.IntType and i >= self._length or i<-self._length+1:
            raise IndexError(i, "is out of range")
        del self._list[i]
        self._length -= oldsize - len(self._list)
    def __repr__(self):
        return repr(self._list[:self._length])
    def __add__(self, other):
        return self._list + other
    def __radd__(self, other):
        return other + self._list

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

if __name__=='__main__':
    l0 = [0, [1,2],[3,[4]], 5]
    l1 = ['zero', 'one', 'two', 'three', 'four', 'five']
    print unflatten(l0, l1)
