# -*- python -*-
# $RCSfile: materialparameter.py,v $
# $Revision: 1.4 $
# $Author: langer $
# $Date: 2006/02/24 16:02:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.


from oof2.common import debug
from oof2.common.IO import parameter
from oof2.engine import materialmanager
import types


# MaterialParameter can be set to any existing Material name.

class MaterialParameter(parameter.StringParameter):
    def checker(self, x):
        if x not in materialmanager.getMaterialNames():
            raise TypeError("Expected a Material name!")

# AnyMaterialParameter can be set to any existing Material name, or
# '<Any>' or '<None>'.

class AnyMaterialParameter(parameter.StringParameter):
    extranames = ['<Any>', '<None>']
    def checker(self, x):
        if x not in materialmanager.getMaterialNames() and x not in self.extranames:
            raise TypeError("Expected a Material name, or '<Any>' or '<None>'")

# ListOfMaterialsParameter can be set to a list of existing Material
# names.

class ListOfMaterialsParameter(parameter.ListOfStringsParameter):
    def checker(self, x):
        if type(x) is not types.ListType:
            raise TypeError("Expected a list of Material names!")
        names = materialmanager.getMaterialNames()
        for n in x:
            if n not in names:
                raise TypeError("Expected a list of Material names!")


