# -*- python -*-
# $RCSfile: microstructuredisplay.py,v $
# $Revision: 1.16 $
# $Author: reida $
# $Date: 2006/05/24 20:14:30 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# DisplayMethod for showing the Materials assigned to a Microstructure.

from oof2.SWIG.common import coord
from oof2.SWIG.engine import material
from oof2.common import color
from oof2.common import debug
from oof2.common import registeredclass
from oof2.common.IO import display
from oof2.common.IO import xmlmenudump
import weakref

# The draw_image() routine requires an object with a .stringimage()
# function.  We can't simply give Microstructure such a function,
# since it can be displayed in many ways, so we construct a temporary
# object, MaterialImage, just to pass to draw_image().

class MaterialImage:
    def __init__(self, microstructure, mmdisplay):
        self.strimage = material.getMaterialImage(microstructure,
                                                  mmdisplay.no_material,
                                                  mmdisplay.no_color)
        self._isize = microstructure.sizeInPixels()
        self._size = microstructure.size()
    def sizeInPixels(self):
        return self._isize
    def size(self):
        return self._size
    def stringimage(self):
        return self.strimage

class MicrostructureMaterialDisplay(display.DisplayMethod):
    def __init__(self, no_material, no_color):
        self.no_material = no_material    # color if Material isn't assigned
        self.no_color = no_color    # color if Material has no ColorProperty
        display.DisplayMethod.__init__(self)
        self.image = weakref.WeakKeyDictionary()
    def draw(self, gfxwindow, device):
        debug.fmsg()
        microstructure = self.who().getObject(gfxwindow)
        debug.fmsg("Getting material image.")
        matlimage = MaterialImage(microstructure, self)
        debug.fmsg("Back.")
        device.draw_image(matlimage, coord.Coord(0,0), microstructure.size())
        # The canvas assumes that its stringimage is persistent, so we
        # have to keep a reference to matlimage here, since matlimage
        # wraps the stringimage.  matlimage can be disposed of when
        # the device goes away or this layer is destroyed, so we use a
        # weak key dictionary.
        self.image[device] = matlimage
        debug.fmsg("Exiting.")
    def getTimeStamp(self, gfxwindow):
        microstructure = self.who().getObject(gfxwindow)
        return max(display.DisplayMethod.getTimeStamp(self, gfxwindow),
                   material.getMaterialTimeStamp(microstructure))

registeredclass.Registration(
    'Material',
    display.DisplayMethod,
    MicrostructureMaterialDisplay,
    ordering=0,
    layerordering=display.Planar,
    params=[
    color.ColorParameter('no_material', color.black,
                         tip="Color to use if no material has been assigned to a pixel"),
    color.ColorParameter('no_color', color.blue,
                         tip="Color to use if the assigned material has no assigned color")
    ],
    whoclasses = ('Microstructure',),
    tip="Display the color of the Material assigned to each pixel.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/microstructuredisplay.xml')
    )
