// -*- C++ -*-
// $RCSfile: bdyanalysis.C,v $
// $Revision: 1.1 $
// $Author: langer $
// $Date: 2006/07/25 20:20:39 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "engine/edge.h"
#include "engine/edgeset.h"
#include "engine/element.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/field.h"
#include "engine/outputval.h"
#include "engine/gausspoint.h"
#include "engine/matvec.h"
#include "engine/SparseLib++/mvblasd.h"


// Helper function for integrating the flux through a boundary. 
VECTOR_D *integrateFlux(const FEMesh *m, const Flux *flux, const EdgeSet *es) {

  VECTOR_D *result = new VECTOR_D(flux->divergence_dim(),0.0);

  for(EdgeSetIterator esi = EdgeSetIterator(es); !esi.end(); ++esi) {

    BoundaryEdge *bdy_edge = esi.edge();
    const Element *el = bdy_edge->el;
    int order = el->shapefun_degree();

    for(EdgeGaussPoint egpt=bdy_edge->integrator(order); !egpt.end(); ++egpt) {

      VECTOR_D *res = flux->contract(m, el, egpt);

      // TODO LATER: Accumulation line makes BLAS calls to += and *
      // operators, might be effort overkill for such small vectors.
      (*result) += (*res)*egpt.weight();
      delete res;
    }
  }
  return result;
}

OutputValue averageField(const FEMesh *m, const Field *field, const EdgeSet *es)
{
  OutputValue result(field->newOutputValue());
  double weight = 0.0;
  for(EdgeSetIterator esi=EdgeSetIterator(es); !esi.end(); ++esi) {
    BoundaryEdge *bdy_edge = esi.edge();
    const Element *el = bdy_edge->el;
    int order = el->shapefun_degree();

    for(EdgeGaussPoint egpt=bdy_edge->integrator(order); !egpt.end(); ++egpt) {
      double w = egpt.weight();
      result += el->outputField(m, *field, egpt.mastercoord())*w;
      weight += w;
    }
  }
  return result/weight;
}
