// -*- C++ -*-
// $RCSfile: cstrain.C,v $
// $Revision: 1.10 $
// $Author: langer $
// $Date: 2005/12/27 22:13:49 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "engine/cstrain.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/field.h"
#include "engine/mastercoord.h"
#include "engine/symmmatrix.h"


// Add the local geometrical strain to the given SymmMatrix.

void findGeometricStrain(const FEMesh *mesh, const Element *element,
			 const MasterPosition &pos, SymmMatrix &strain)
{
  static CompoundField *displacement =
    dynamic_cast<CompoundField*>(Field::getField("Displacement"));;
  static Field *displacement_z = displacement->out_of_plane();

  if(!mesh->is_defined_field(*displacement))
    return;

  bool inplane = displacement->in_plane(mesh);
  
  for(ElementFuncNodeIterator nu=element->funcnode_iterator(); !nu.end(); ++nu) 
    {
      // Shape function derivatives at postion pos.
      double dsfx = nu.dshapefunction(0, pos);
      double dsfy = nu.dshapefunction(1, pos);
      // Displacement components at node nu.
      double dispx = (*displacement)(mesh, nu, 0)->value();
      double dispy = (*displacement)(mesh, nu, 1)->value();
      strain(0,0) += dispx*dsfx;
      strain(1,1) += dispy*dsfy;
      strain(0,1) += 0.5*(dispx*dsfy + dispy*dsfx);

      if(!inplane) {
	double sf = nu.shapefunction(pos);
	strain(0,2) += sf*(*displacement_z)(mesh, nu, 0)->value();
	strain(1,2) += sf*(*displacement_z)(mesh, nu, 1)->value();
	strain(2,2) += sf*(*displacement_z)(mesh, nu, 2)->value();
      }
    }
}


// Initialize a PropertyOutput with the geometrical strain.

OutputVal *POInitGeometricStrain::operator()(const PropertyOutput *po,
					     const FEMesh *mesh,
					     const Element *element,
					     const MasterCoord &pos) const
{
  SymmMatrix3 *strain = new SymmMatrix3();
  // Get the Python parameter that tells which Strain is being
  // computed. It's a RegisteredParameter.
  std::string straintype = po->getRegisteredParamName("type");
  if(straintype == "Geometric" || straintype == "Elastic") {
    findGeometricStrain(mesh, element, pos, *strain);
  }
  return strain;
};
