// -*- C++ -*-
// $RCSfile: stressfreestrain.h,v $
// $Revision: 1.2 $
// $Author: langer $
// $Date: 2006/07/12 15:04:19 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef STRESSFREESTRAIN_H
#define STRESSFREESTRAIN_H

#include <oofconfig.h>

#include "engine/property.h"
#include "engine/symmmatrix.h"
#include <string>

class Elasticity;
class Element;
class Flux;
class FluxData;
class Material;
class OrientationProp;
class SymmetricTensorFlux;

class StressFreeStrain : public Property {
protected:
  Elasticity *elasticity;
  SymmetricTensorFlux *stress_flux;
  SymmMatrix3 stressfreestrain;
public:
  StressFreeStrain(PyObject *registry,
	      const std::string &nm);
  virtual ~StressFreeStrain() {}
  virtual void cross_reference(Material*) = 0;

  virtual void fluxrhs(const FEMesh*, const Element*,
		       const Flux*, FluxData*, const MasterPosition&) const;

  virtual void output(const FEMesh*, const Element*, const PropertyOutput*,
		      const MasterPosition&, OutputVal*) const;

  virtual int integration_order(const FEMesh*, const Element*) const;
  virtual bool is_symmetric(const FEMesh*) const { return true; }
};

class IsotropicStressFreeStrain: public StressFreeStrain {
public:
  IsotropicStressFreeStrain(PyObject *registry,
		       const std::string &name,
		       double e);
  virtual void cross_reference(Material*);
  virtual void precompute();
private:
  double e_;
};

class AnisotropicStressFreeStrain: public StressFreeStrain {
public:
  AnisotropicStressFreeStrain(PyObject *registry,
			 const std::string &nm,
			 SymmMatrix3 *e);
  virtual void cross_reference(Material*);
  virtual void precompute();
private:
  SymmMatrix3 e_;
  OrientationProp *orientation;
};

#endif
