// -*- C++ -*-
// $RCSfile: pypropertywrapper.h,v $
// $Revision: 1.29 $
// $Author: langer $
// $Date: 2006/03/15 19:44:27 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef PYPROPERTYWRAPPER_H
#define PYPROPERTYWRAPPER_H

// Structure to wrap a pure-Python property, and make it available by
// way of the usual interface to the stiffness-matrix construction
// process.

#include <Python.h>
#include "engine/element.h"
#include "property.h"
#include "common/pythonexportable.h"

class Field;
class Flux;
class Material;
class FEMesh;
class ElementFuncNodeIterator;
class MasterPosition;

// Data storage class for objects descended from PyPropertyWrapper.
class PyPropertyElementData : public ElementData {
private:
  PyObject *_data;
  static std::string classname_;
  static std::string modulename_;
public:
  PyPropertyElementData(const std::string &name, PyObject *dat);
  virtual ~PyPropertyElementData();
  virtual const std::string &classname() const { return classname_; }
  virtual const std::string &modulename() const { return modulename_; }
  PyObject *data();
};

class PyPropertyWrapper : public Property, 
			  virtual public PythonNative<Property> {
public:
  PyPropertyWrapper(PyObject *referent, PyObject *regstn,
		    const std::string &name);
  virtual ~PyPropertyWrapper();

  // The following functions use the Python/C API to call Python
  // functions defined in the referent object.  The functions can all
  // be optionally redefined in the Python subclass.  If they're not
  // redefined, the default Property base class function will be used.

  // This is only necessary for functions defined in Python and called
  // from C++.  Functions defined in Python and called from Python
  // (such as is_symmetric()) can simply be defined in the Python
  // subclass and don't need to be mentioned here at all.

  // TODO: It's not clear what to do about functions that might need
  // to be redefined in Python and are called from both C++ and
  // Python.  The mechanism used for the functions below leads to
  // infinite recursion when called from Python when the base class
  // function has *not* been redefined in Python.

  virtual void precompute();
  virtual void cross_reference(Material*);
  virtual void begin_element(const Element*);
  virtual void end_element(const Element*);
  virtual void post_process(FEMesh *, const Element*) const;
  virtual int integration_order(const FEMesh*, const Element*) const;
  virtual void fluxmatrix(const FEMesh*, const Element*,
			  const ElementFuncNodeIterator&,
			  const Flux*, FluxData*,
			  const MasterPosition&) const;
  virtual void fluxrhs(const FEMesh*, const Element*, const Flux*, FluxData*,
		       const MasterPosition&) const;

  virtual void output(const FEMesh*, const Element*, const PropertyOutput*,
		      const MasterPosition&, OutputVal*) const;

//   // The PyPropertyWrapper also needs to provide an interface to the 
//   // element's data storage and retrieval.
//   int appendElementData(Element *el, PyPropertyElementData *data);
//   PyPropertyElementData *getElementData(Element *el, int i);
//   int getElementDataIndex(Element *el, std::string name);

private:
  PyObject *referent_;      // Pointer to the actual Python object.
};

#endif // PYPROPERTYWRAPPER_H
